#ifndef M6502_H
#define M6502_H

// See cpu.cc for copyrights.

#include "defs.h"
                               /* 6502 status flags:         */
#define	C_FLAG	  0x01         /* 1: Carry occured           */
#define	Z_FLAG	  0x02         /* 1: Result is zero          */
#define	I_FLAG	  0x04         /* 1: Interrupts disabled     */
#define	D_FLAG	  0x08         /* 1: Decimal mode            */
#define	B_FLAG	  0x10         /* Break [0 on stk after int] */
#define	R_FLAG	  0x20         /* Always 1                   */
#define	V_FLAG	  0x40         /* 1: Overflow occured        */
#define	N_FLAG	  0x80         /* 1: Result is negative      */

/** Rd6502()/Wr6502/Op6502() *********************************/
/** These functions are called when access to RAM occurs.   **/
/** They allow to control memory access. Op6502 is the same **/
/** as Rd6502, but used to read *opcodes* only, when many   **/
/** checks can be skipped to make it fast. It is only       **/
/** required if there is a #define FAST_RDOP.               **/
/************************************ TO BE WRITTEN BY USER **/
extern void Wr6502(register fWORD Addr, register fBYTE Value);
extern BYTE Rd6502(register fWORD Addr);
extern BYTE Op6502(register fWORD Addr) CONSTFUN;

extern BYTE Clocks[256];
extern BYTE ZNTable[256];
class m6502cpu
{
public:
    BYTE A,P,X,Y,S;
    PAIR PC;

private:
    fWORD ICount;
    bool IRQPending;
private:
    inline BYTE fetchop()
    {
    	register BYTE i = Op6502(PC.W++);
    	ICount += Clocks[i];
    	return i;
    }
    inline WORD M_LDWORD()
    {
		register WORD Rg = Op6502(PC.W++);
		Rg |= Op6502(PC.W++) << 8;
		return Rg;
    }
    
    /** Addressing Methods ***************************************/
    /** These macros calculate and return effective addresses.  **/
    /*************************************************************/
    inline WORD MC_Ab()	{ return M_LDWORD(); }
    inline WORD MC_Zp()	{ return BYTE(Op6502(PC.W++)); }
    inline WORD MC_Zx()	{ return BYTE(Op6502(PC.W++)+X); }
    inline WORD MC_Zy()	{ return BYTE(Op6502(PC.W++)+Y); }
    inline WORD MC_Ax()	{ return M_LDWORD()+X; }
    inline WORD MC_Ay()	{ return M_LDWORD()+Y; }
    inline WORD MC_Ix() { register WORD K = BYTE(Op6502(PC.W++) + X); return Op6502(K) | (Op6502(K+1) << 8); }
    inline WORD MC_Iy() { register WORD K = Op6502(PC.W++); return (Op6502(K) | (Op6502(K+1) << 8)) + Y; }
    /** Reading From Memory **************************************/
    /** These macros calculate address and read from it.        **/
    /*************************************************************/
    inline BYTE MR_Ab()	{ return Rd6502(MC_Ab()); }
    inline BYTE MR_Im()	{ return Op6502(PC.W++); }
    inline BYTE MR_Zp()	{ return Rd6502(MC_Zp()); }
    inline BYTE MR_Zx()	{ return Rd6502(MC_Zx()); }
    inline BYTE MR_Zy()	{ return Rd6502(MC_Zy()); }
    inline BYTE MR_Ax()	{ return Rd6502(MC_Ax()); }
    inline BYTE MR_Ay()	{ return Rd6502(MC_Ay()); }
    inline BYTE MR_Ix()	{ return Rd6502(MC_Ix()); }
    inline BYTE MR_Iy()	{ return Rd6502(MC_Iy()); }
    /** Writing To Memory ****************************************/
    /** These macros calculate address and write to it.         **/
    /*************************************************************/
    inline void MW_Ab(const BYTE r) { Wr6502(MC_Ab(), r); }
    inline void MW_Zp(const BYTE r) { Wr6502(MC_Zp(), r); }
    inline void MW_Zx(const BYTE r) { Wr6502(MC_Zx(), r); }
    inline void MW_Zy(const BYTE r) { Wr6502(MC_Zy(), r); }
    inline void MW_Ax(const BYTE r) { Wr6502(MC_Ax(), r); }
    inline void MW_Ay(const BYTE r) { Wr6502(MC_Ay(), r); }
    inline void MW_Ix(const BYTE r) { Wr6502(MC_Ix(), r); }
    inline void MW_Iy(const BYTE r) { Wr6502(MC_Iy(), r); }
    
    inline BYTE M_POP() { return Op6502(++S | 0x100); }
    inline void M_PUSH(const BYTE r) { Wr6502(0x100|S--, r); }
    inline void M_FL(const BYTE r) { P = (P & ~(Z_FLAG|N_FLAG)) | ZNTable[r]; }
    
    inline void M_AND(const BYTE r) { M_FL(A &= r); }
    inline void M_ORA(const BYTE r) { M_FL(A |= r); }
    inline void M_EOR(const BYTE r) { M_FL(A ^= r); }
    inline void M_SBC(const BYTE r)
    {
        /* Warning! C_FLAG is inverted before SBC and after it */
        PAIR K;
        /* NES does not use this
        if(P&D_FLAG)
        {
            K.B.l=(A&0x0F)-(r&0x0F)-(~P&C_FLAG);
            if(K.B.l&0x10) K.B.l-=6;
            K.B.h=(A>>4)-(r>>4)-(K.B.l&0x10);
            if(K.B.h&0x10) K.B.h-=6;
            A=(K.B.l&0x0F)|(K.B.h<<4);
            P=(P&~C_FLAG)|(K.B.h>15? 0:C_FLAG);
        }
        else
        {*/
            K.W=A-r-(~P&C_FLAG);
            P&=~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);
            P|=((A^r)&(A^K.B.l)&0x80? V_FLAG:0)|
               (K.B.h? 0:C_FLAG)|ZNTable[K.B.l];
            A=K.B.l;
        /*}*/
    }
    inline void M_ADC(const BYTE r)
    {
        PAIR K;
        /* NES does not use this
        if(P&D_FLAG)
        {
            K.B.l=(A&0x0F)+(r&0x0F)+(P&C_FLAG);
            K.B.h=(A>>4)+(r>>4)+(K.B.l>15? 1:0);
            if(K.B.l>9) { K.B.l+=6;++K.B.h; }
            if(K.B.h>9) K.B.h+=6;
            A=(K.B.l&0x0F)|(K.B.h<<4);
            P=(P&~C_FLAG)|(K.B.h>15? C_FLAG:0);
        }
        else
        { */
            K.W = A + r + (P & C_FLAG);
            P &= ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG);
            P |= (~(A^r)&(A^K.B.l)&0x80? V_FLAG:0) | (K.B.h?C_FLAG:0) | ZNTable[K.B.l];
            A = K.B.l;
        /*}*/
    }
    inline void M_BIT(const BYTE r)
    {
    	P = (P & ~(N_FLAG|V_FLAG|Z_FLAG)) | (r&(N_FLAG|V_FLAG)) | (r&A? 0:Z_FLAG);
	}
	inline void M_CMP(const BYTE r1, const BYTE r2)
	{
        PAIR K;
        K.W = r1 - r2;
        P &= ~(N_FLAG|Z_FLAG|C_FLAG);
        P |= ZNTable[K.B.l] | (K.B.h? 0:C_FLAG);
	}
public:
	m6502cpu() : ICount(0), IRQPending(false)
	{
		// Do not reset here, as the memory
		// has probably not yet been initialized.
	}
    /** Reset6502() **********************************************/
    /** This function can be used to reset the registers before **/
    /** starting execution with Run6502(). It sets registers to **/
    /** their initial values.                                   **/
    /*************************************************************/
	void Reset()
	{
        A = X = Y = 0x00;
        P = Z_FLAG|R_FLAG;
        S = 0xFF;
        PC.B.l = Op6502(0xFFFC);
        PC.B.h = Op6502(0xFFFD);
        ICount = 0;
        IRQPending = false;
	}
	
    /** Int6502() ************************************************/
    /** This function will generate interrupt of a given type.  **/
    /** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
    /** will cause a normal interrupt, unless I_FLAG set in R.  **/
    /*************************************************************/
    void Int(InterruptType Type);

    void Tick();
    
    inline bool ready() { if(!ICount)return true; --ICount; return false; }
};

#endif
